# !/usr/bin/env Rscript
# 11/10/2017

# @title Single-cell Aggregated clustering From Ensemble (SAFE)
#
# @description This function performs single-cell clustering using four state-of-the-art methods, 
# SC3, CIDR, Seurat and tSNE+kmeans.
# 
# @param inputTags a G*N matrix with G genes and N cells.
# @param datatype defines the type of data, which could be "count", "CPM", "RPKM" and "FPKM". 
# Default is "count".
# @param SC3 a boolean variable that defines whether to cluster cells using SC3 method. 
# Default is "TRUE".
# @param gene_filter a boolean variable defines whether to perform gene filtering 
# before SC3 clustering, when \code{SC3 = TRUE}.
# @param svm_max, if \code{SC3 = TRUE}, then defines the mimimum number of cells above which SVM will be run.
# @param CIDR a boolean parameter that defines whether to cluster cells using CIDR method. 
# Default is "TRUE".
# @param Seurat is a boolean variable that defines whether to cluster cells using Seurat method. 
# Default is "TRUE".
# @param nPC defines the number of princple compoents used in Seurat clustering, when \code{Seurat = TRUE}. 
# Default value is esimated by \code{nPC} of \code{CIDR}.
# @param resolution defines the value of resolution used in Seurat clustering, when \code{Seurat = TRUE}.
# @param seurat_min_cell defines the mimimum number of cells in input dataset below which 
# \code{resolution} is set to 1.2, when \code{Seurat = TRUE}.
# @param resolution_min defines the resolution used in Seurat clustering for small dataset, 
# when \code{Seurat == TRUE} and cell number of input file < \code{seurat_min_cell}.
# @param tSNE is a boolean variable that defines whether to cluster cells using tSNE method.
# Default is "TRUE".
# @param var_genes defines the number of variable genes used by tSNE analysis, when \code{tSNE = TRUE}.
# @param SEED sets the seed of the random number generator. Setting the seed to a fixed value can 
# produce reproducible clustering results. 
#
# @return a matrix of indiviudal clustering results is returned.
#
# @author Yuchen Yang <yyuchen@email.unc.edu>, Ruth Huh <rhuh@live.unc.edu>, 
# Houston Culpepper <hculpepp@live.unc.edu>, Yun Li <yunli@med.unc.edu>
# @citation Yuchen Yang, Ruth Huh, Houston Culpepper, Yun Li. SAFE (Single-cell Aggregated clustering From Ensemble): Cluster ensemble for single-cell RNA-seq data. 2017

individual_clustering <- function(inputTags, datatype = "count", 
                            SC3 = TRUE, gene_filter = FALSE, svm_max = 5000, CIDR = TRUE, 
                            Seurat = TRUE, nPC = NULL, resolution = 0.9, seurat_min_cell = 200, resolution_min = 1.2, 
                            tSNE = TRUE, var_genes = NULL, SEED = 1){
  
  set.seed(SEED)
  
  
  library(cidr)  #https://github.com/VCCRI/CIDR
  library(scater)  #https://github.com/davismcc/scater
  library(SC3)  #https://github.com/hemberg-lab/SC3
  library(e1071)
  library(Seurat)  #http://satijalab.org/seurat/get_started.html
  library(dplyr)
  library(Matrix)
  library(Rtsne)
  library(ADPclust)
  
  input_lcpm <- NULL
  input_ltpm <- NULL
  sc3OUTPUT <- NULL
  cidrOUTPUT <- NULL
  seuratOUTPUT <- NULL
  tsne_input <- NULL
  tsne_output <- NULL
  tsne_kmeansOUTPUT <- NULL
  cluster_results <- NULL
  adpOUTPUT <- NULL
  cluster_results <- NULL
  
  inputTags = as.matrix(inputTags)
  ##### SC3
  if(SC3 == TRUE){
    exp_cell_info <- NULL
    cell_inds <-NULL
    pd <- NULL
    exp_cell_exprs <- NULL
    exp_sceset <- NULL
    
    # cell annotation
    exp_cell_info = data.frame(cell_id = colnames(inputTags))
    cell_inds <- paste("Cell", 1:ncol(inputTags), sep = "_")
    rownames(exp_cell_info) <- cell_inds
    pd <- new("AnnotatedDataFrame", data = exp_cell_info)
    
    # cell expression
    if (datatype == "count" || datatype == "CPM") {
      ### If the input data is original count data or CPM, it can be used directly
      exp_cell_exprs <- inputTags
    } else if (datatype == "FPKM" || datatype == "RPKM") {
      ### If the input data is FPKM or RPKM, it would be transformed to TPM
      input_tpm = t(t(inputTags)/colSums(inputTags))*1000000
      exp_cell_exprs <- input_tpm
    }
    
    colnames(exp_cell_exprs) <- cell_inds
    
    # SCESEt object
    exp_cell_exprs <- as.matrix(exp_cell_exprs)
    
    if (datatype == "count") {
      ### If the input data is original count data, it can be used directly
      exp_sceset <- newSCESet(countData = exp_cell_exprs, phenoData = pd)
    } else if (datatype == "CPM") {
      ### If the input data is CPM, it can be used directly
      exp_sceset <- newSCESet(cpmData = exp_cell_exprs, phenoData = pd)
    } else if (datatype == "FPKM" || datatype == "RPKM") {
      ### If the input data is FPKM or RPKM, the transformed TPM would be used as input
      exp_sceset <- newSCESet(tpmData = exp_cell_exprs, phenoData = pd)
    }
    
    is_exprs(exp_sceset) <- exprs(exp_sceset) > 0
    exp_sceset <- calculateQCMetrics(exp_sceset)
    exp_sceset <- sc3_prepare(exp_sceset, ks = 2, gene_filter = gene_filter, rand_seed = SEED)
    
    ### Estimating optimal number of clustering
    exp_sceset <- sc3_estimate_k(exp_sceset)
    optimal_K = exp_sceset@sc3$k_estimation
    
    ### Clustering by SC3 at the optimal K
    exp_sceset <- sc3(exp_sceset, ks = optimal_K, biology = FALSE, gene_filter = gene_filter, svm_max = svm_max, rand_seed = SEED)
    
    ### Support vector machines (SVM)
    support_vector_machines <- function(train, study, kern) {
      train <- t(train)
      labs <- factor(rownames(train))
      rownames(train) <- NULL
      model <- tryCatch(svm(train, labs, kernel = kern), error = function(cond) return(NA))
      pred <- predict(model, t(study))
      return(pred = pred)
    }
    
    sc3_run_svm.SCESet <- function(object) {
      if (is.null(object@sc3$svm_train_inds)) {
        warning(paste0("Please rerun sc3_prepare() defining the training cells!"))
        return(object)
      }
      
      dataset <- get_processed_dataset(object)
      ks <- object@sc3$ks
      p_data <- object@phenoData@data
      svm_train_inds <- object@sc3$svm_train_inds
      svm_study_inds <- object@sc3$svm_study_inds
      
      for (k in ks) {
        clusts <- p_data[, paste0("sc3_", k, "_clusters")]
        clusts <- clusts[svm_train_inds]
        
        train.dataset <- dataset[, svm_train_inds]
        colnames(train.dataset) <- clusts
        
        study.labs <- support_vector_machines(train.dataset, dataset[, svm_study_inds], "linear")
        svm.labs <- c(clusts, study.labs)
        ord <- order(c(svm_train_inds, svm_study_inds))
        
        p_data[, paste0("sc3_", k, "_clusters")] <- svm.labs[ord]
      }
      pData(object) <- new("AnnotatedDataFrame", data = p_data)
      return(object)
    }
    
    ### Runing SVM
    if (ncol(inputTags) >= svm_max){
      exp_sceset <- sc3_run_svm.SCESet(exp_sceset)
    }
    
    ### Exporting SC3 results
    p_Data <- pData(exp_sceset)
    col_name <- paste("sc3_", optimal_K, "_clusters", sep = '')
    sc3OUTPUT <- p_Data[, grep(col_name, colnames(p_Data))]
    cluster_results = rbind(cluster_results, matrix(c(sc3OUTPUT), nrow = 1, byrow = T))
  }
  
  
  if(CIDR == TRUE){
    ##### CIDR
    cidrOUTPUT <- scDataConstructor(inputTags)
    cidrOUTPUT <- determineDropoutCandidates(cidrOUTPUT)
    cidrOUTPUT <- wThreshold(cidrOUTPUT)
    cidrOUTPUT <- scDissim(cidrOUTPUT)
    cidrOUTPUT <- scPCA(cidrOUTPUT)
    cidrOUTPUT <- nPC(cidrOUTPUT)
    
    ### nPC is not necessarily the same between CIDR and Seurat. Here we use the same nPC value for both two methods for convenience
    if(!is.null(nPC)) {
      cidrOUTPUT@nPC <- nPC
    } else {
      nPC <- cidrOUTPUT@nPC
    }
    
    ### Clustering by CIDR.
    # The optimal clustering number is determined automatically
    cidrOUTPUT <- scCluster(cidrOUTPUT, nPC = nPC)
    cluster_results = rbind(cluster_results, matrix(c(cidrOUTPUT@clusters), nrow = 1, byrow = T))
  }
  
  
  if(Seurat == TRUE){
    ##### Seurat
    # Initialize the Seurat object with the raw data (non-normalized data)
    seuratOUTPUT <- new("seurat", raw.data = inputTags)
    
    # Keep all genes expressed in >= 3 cells, keep all cells with >= 200 genes
    # Perform log-normalization, first scaling each cell to a total of 1e4 molecules (as in Macosko et al. Cell 2015)
    seuratOUTPUT <- Setup(seuratOUTPUT, min.cells = 3, min.genes = 200, do.logNormalize = T, total.expr = 1e4, project = "single-cell clustering")
    
    # Regress out unwanted sources of variation
    seuratOUTPUT <- RegressOut(seuratOUTPUT, latent.vars = c("nUMI"))
    
    # Detection of variable genes across the single cells
    seuratOUTPUT <- MeanVarPlot(seuratOUTPUT,fxn.x = expMean, fxn.y = logVarDivMean, x.low.cutoff = 0.0125, x.high.cutoff = 3, y.cutoff = 0.5, do.plot = FALSE, do.contour = F)
    
    ### Perform linear dimensional reduction
    seuratOUTPUT <- PCA(seuratOUTPUT, pc.genes = seuratOUTPUT@var.genes, do.print = FALSE, pcs.print = 5, genes.print = 5)
    
    if (length(inputTags[1,]) >= seurat_min_cell){
      ### Determine statistically significant principal components
      # NOTE: This process can take a long time for big datasets, comment out for expediency.
      # More approximate techniques such as those implemented in PCElbowPlot() can be used to reduce computation time
      # Here we chooes the same number of PCs used in CIDR
      
      ### Clustering the cells by Seurat
      seuratOUTPUT <- FindClusters(seuratOUTPUT, pc.use = 1:nPC, algorithm = 3, resolution = resolution, print.output = F, random.seed = SEED)
    } else {
      resolution <- resolution_min
      seuratOUTPUT <- FindClusters(seuratOUTPUT, pc.use = 1:nPC, algorithm = 3, resolution = resolution_min, print.output = F, random.seed = SEED)
    }
    
    ### Complementing the missing data
    cells_dropout <- NULL
    num_genes <- colSums(seuratOUTPUT@raw.data > 0)
    cells_dropout <- names(num_genes[which(num_genes <= 200)])
    if (length(cells_dropout != 0)){
      seurat_output <- matrix(NA, ncol = ncol(seuratOUTPUT@raw.data), byrow = T)
      colnames(seurat_output) <- colnames(seuratOUTPUT@raw.data)
      seurat_retained <- t(as.matrix(as.numeric(seuratOUTPUT@ident)))
      colnames(seurat_retained) <- colnames(seuratOUTPUT@data)
      for (i in 1:ncol(seurat_retained)){
        seurat_output[1,colnames(seurat_retained)[i]] <- seurat_retained[1,colnames(seurat_retained)[i]]
      }
    } else {
      seurat_output <- t(as.matrix(as.numeric(seuratOUTPUT@ident)))
    }
    
    cluster_results = rbind(cluster_results, matrix(c(as.numeric(seuratOUTPUT@ident)), nrow = 1, byrow = T))
  }
  
  
  if(tSNE == TRUE){
    ##### tSNE+kmeans
    ### Data tranformation
    if (datatype == "count" || datatype == "CPM") {
      ### If the input data is original count data or CPM, it would be tranformed to CPM
      input_lcpm <- log2(t(t(inputTags)/colSums(inputTags))*1000000+1)
      tsne_input <- input_lcpm
    } else if (datatype == "FPKM" || datatype == "RPKM") {
      ### If the input data is FPKM or RPKM, we use the transformed TPM data generated before as the input
      input_ltpm = log2(t(t(inputTags)/colSums(inputTags))*1000000+1)
      tsne_input <- input_ltpm
    }
    
    ### Dimensionality reduction by Rtsne
    if(length(inputTags[1,]) >= 200) {
      perplexity = 30
    } else {
      perplexity = 10
    }
    
    if (is.null(var_genes)){
      tsne_output <- Rtsne(t(tsne_input), perplexity = perplexity)
    } else{
      se_genes = rep(NA, nrow(tsne_input))
      for (i in 1:nrow(tsne_input)){
        se_genes[i] = sqrt(var(tsne_input[i,])/length(tsne_input[i,]))
      }
      decreasing_rank = order(se_genes, decreasing = T)
      
      tsne_output <- Rtsne(t(tsne_input[decreasing_rank[1:var_genes],]), perplexity = perplexity)
    }
    
    ### Determining the optimal cluster number (k) by ADPclust
    
    mim_K = 2
    max_K = max(optimal_K, cidrOUTPUT@nCluster, max(as.numeric(seuratOUTPUT@ident)))
    
    adpOUTPUT <- adpclust(tsne_output$Y, htype = "amise",centroids="auto", nclust = mim_K:max_K)
    
    ### Clustering the cells by kmeans
    tsne_kmeansOUTPUT <- kmeans(tsne_output$Y, tsne_output$Y[adpOUTPUT$centers,], adpOUTPUT$nclust)
    cluster_results = rbind(cluster_results, matrix(c(tsne_kmeansOUTPUT$cluster), nrow = 1, byrow = T))
  }
  
  #cluster_results = matrix(c(sc3OUTPUT, cidrOUTPUT@clusters, as.numeric(seuratOUTPUT@ident), tsne_kmeansOUTPUT$cluster), nrow = 4, byrow = T)
  
  return(cluster_results)
}
